"""Asian Development Bank (ADB) provider."""

from collections import defaultdict
from itertools import chain, product
from typing import Callable, Iterable, Tuple
from urllib.parse import quote, urlparse

import pandas as pd
import sdmx.model.v21 as m
from sdmx.model import v21

from transport_data.util.pluggy import hookimpl
from transport_data.util.pooch import Pooch
from transport_data.util.sdmx import anno_generated

BASE_URL = "https://asiantransportoutlook.com/exportdl?orig=1"

#: List of all "ECONOMY" codes appearing in processed data.
CL_ECONOMY = m.Codelist(
    id="ECONOMY",
    name="Asian Transport Outlook subject economy",
    description="Equivalent to GEO or REF_AREA in other concept schemes.",
)

#: List of all measures (indicators) appearing in processed data.
#:
#: .. todo:: Validate against the master list of indicators; or read from that file and
#:    validate IDs appearing in data files.
CS_MEASURE = m.ConceptScheme(
    id="MEASURE",
    name="Asian Transport Outlook measures (indicators)",
    description="Item are automatically generated by the TDCI tools. Currently no "
    "correspondence with the list provided directly by ADB is checked or enforced.",
)

#: Mapping from short codes for ATO data categories to file names.
FILES = {
    # "ATO National Database Masterlist of Indicators",
    "ACC": (
        "ATO Workbook (ACCESS & CONNECTIVITY (ACC)).xlsx",
        "sha256:21c3b7e662d932f5cc61c22489acb3cf0e8a70200abc2372c7fe212602903fd7",
    ),
    "APH": (
        "ATO Workbook (AIR POLLUTION & HEALTH (APH)).xlsx",
        "sha256:b06c102a1184ed83d77673146599e11c2b6c81d784ebcee6b46f4d43713e899a",
    ),
    "CLC": (
        "ATO Workbook (CLIMATE CHANGE (CLC)).xlsx",
        "sha256:7fe2d4bb656508bf3194406d25ed04c1ac403daaaf1ada74847a2c330efcfb2a",
    ),
    "INF": (
        "ATO Workbook (INFRASTRUCTURE (INF)).xlsx",
        "sha256:30b9d05330838809ff0d53b2e61ade935eccdb4b73c0509fd1fc49b405699ac3",
    ),
    "MIS": (
        "ATO Workbook (MISCELLANEOUS (MIS)).xlsx",
        "sha256:2ef3cdc5e6363cdca1f671bbf12bf0463fe8a4210cb49b3d32ebd2440c6fe6df",
    ),
    "POL": (
        "ATO Workbook (TRANSPORT POLICY (POL)).xlsx",
        "sha256:fbf23b012590b631239654d255d23ccb70fa717b466be8343a5b0f1e8b4ce720",
    ),
    "RSA": (
        "ATO Workbook (ROAD SAFETY (RSA)).xlsx",
        "sha256:a4285d129c2739c8660a07a5d1c9902ec21c3cd4d13a2a9dfe9d49daca2c0dd5",
    ),
    "SEC": (
        "ATO Workbook (SOCIO-ECONOMIC (SEC)).xlsx",
        "sha256:b5d2ee5d07b5554ef262436ef898afd976fbb4956bd7cb850829cb7391d207c0",
    ),
    "TAS": (
        "ATO Workbook (TRANSPORT ACTIVITY & SERVICES (TAS)).xlsx",
        "sha256:628d4e9774f84d30d80706e982e4ddf7187d77f8676e69765c307701da1caf77",
    ),
}

VERSION = "0.1.0"


def expand(fname: str) -> str:
    return FILES.get(fname, (fname, None))[0]


def _make_url(fname: str) -> str:
    return f"{BASE_URL}&filename={quote(expand(fname))}"


POOCH = Pooch(
    module=__name__,
    base_url=BASE_URL,
    registry={v[0]: v[1] for v in FILES.values()},
    urls={expand(key): _make_url(key) for key in FILES.keys()},
    expand=expand,
)


def convert(part: str):
    from tqdm import tqdm

    from transport_data import STORE

    path = POOCH.fetch(part)
    ef = pd.ExcelFile(path, engine="openpyxl")

    # Maximum length of a sheet name, for formatting the description
    L = max(map(lambda v: len(str(v)), ef.sheet_names))

    for sheet_name in (progress := tqdm(ef.sheet_names)):
        if sheet_name == "TOC":
            # Skip the TOC
            continue
        progress.set_description(f"Process sheet {sheet_name!r:>{L}}")

        df, annos = read_sheet(ef, str(sheet_name))
        ds = convert_sheet(df, annos)

        # Write the DSD and DFD
        STORE.set(ds.described_by)
        STORE.set(ds.structured_by)
        # Write the data itself, to SDMX-ML and CSV
        STORE.set(ds)

    # Write the lists of "Economy" codes and measures/concepts accumulated while
    # converting
    a = get_agencies()[0]
    for obj in (CL_ECONOMY, CS_MEASURE):
        obj.maintainer = a
        obj.version = VERSION
        STORE.set(obj)


def convert_sheet(df: pd.DataFrame, aa: m.AnnotableArtefact):
    """Convert `df` and `aa` from :func:`read_sheet` into SDMX data structures."""
    # Prepare an empty data set, associated structures, and a helper function
    ds, _make_obs = prepare(aa)

    # - Validate values in "Economy Code", "Economy Name" against CL_ECONOMY; then keep
    #   only the former as "ECONOMY".
    # - Melt into long format (one observation per row), preserving remarks columns.
    # - Drop NA values in the "OBS_VALUE" column
    data = (
        df.pipe(validate_economy)
        .melt(
            id_vars=["ECONOMY"] + (ds.eval_annotation("remark-cols") or []),
            var_name="TIME_PERIOD",
            value_name="OBS_VALUE",
        )
        .dropna(subset=["OBS_VALUE"])
    )

    # Convert rows of `data` into SDMX Observation objects
    ds.obs.extend(_make_obs(row) for _, row in data.iterrows())
    assert len(ds) == len(data)

    ds.pop_annotation(id="remark-cols")

    return ds


ATTRIBUTES: set[str] = set()


def _is_url(value: str) -> bool:
    """Return :any:`True` if `value` contains a URL."""
    return urlparse(value).path != value


def dataset_to_metadata_reports(
    ds: "v21.DataSet", msd: "v21.MetadataStructureDefinition"
) -> Iterable["v21.MetadataReport"]:
    """Convert the attributes of ATO `ds` to 1 or more :class:`.MetadataReport`.

    The metadata reports conform to the TDC metadata structure (
    :func:`.org.metadata.get_msd`).

    If `ds` contains per-series values for attributes named “Source”, “Source
    (2024-11)”, or similar, then additional metadata reports are generated, one for each
    series (=GEO, or ‘economy’) and each distinct upstream source indicated by these
    attribute values.
    """
    import sdmx.urn
    from pycountry import countries

    from transport_data import STORE
    from transport_data.org.metadata import make_ra, make_tok

    # Retrieve the DSD and DFD
    assert ds.structured_by and ds.structured_by.urn
    urn_dsd = sdmx.urn.normalize(ds.structured_by.urn)
    dsd = STORE.get(urn_dsd)
    dfd = STORE.get(urn_dsd.replace("DataStructure", "Dataflow"))
    # Assign the loaded DSD to the DFD's `structure` attribute, as these are loaded
    # separately
    dfd.structure = dsd

    # - Create a TargetObjectKey to associate the metadata report with the data flow
    #   definition
    # - Create the report itself
    mdr = v21.MetadataReport(attaches_to=make_tok(dfd))

    # Add the "DATAFLOW" metadata attribute, containing the same information
    mdr.metadata.append(make_ra("DATAFLOW", f"ATO:{dfd.id}"))

    # Map:
    # - from ATO IDs for attributes attached to data sets
    # - to TDC IDs for metadata attributes
    for attribute_id, mda_id in (
        ("SOURCE", "DATA_PROVIDER"),
        ("UNITS", "UNIT_MEASURE"),
        ("WEBSITE", "URL"),
    ):
        # Retrieve the (data) attribute value
        value = str(ds.attrib[attribute_id].value)
        # Convert to a reported (metadata) attribute; append to the report
        value = format_data_provider(value) if mda_id == "DATA_PROVIDER" else value
        mdr.metadata.append(make_ra(mda_id, value))

    # Retrieve the MEASURE from CS_MEASURE
    cs_measure = STORE.get("ConceptScheme=ADB:MEASURE(0.1.0)")
    measure_concept = cs_measure[dsd.id]
    # Construct a metadata attribute
    mdr.metadata.append(make_ra("MEASURE", measure_concept.name))

    # Assemble description
    description_lines = [
        f"Original (ATO) description of {dsd.id!r}:",
        str(measure_concept.description),
    ]

    ra_comment = make_ra("COMMENT", "Metadata generated by transport_data.tools.")
    mdr.metadata.append(ra_comment)

    geo_all = set()  # All GEO/REF_AREA for which the data set contains data
    sources = defaultdict(set)  # Mapping from GEO → "Source…" attribute values

    for observation in ds.obs:
        # Identify GEO/REF_AREA for which the data set contains data
        # Key value for the 'ECONOMY' dimension of `observation`
        geo_alpha_3 = observation.key["ECONOMY"].value
        # Convert from ISO 3166-1 alpha-3 (used by ATO) to alpha-2 (SDMX convention)
        geo_alpha_2 = countries.lookup(geo_alpha_3).alpha_2
        geo_all.add(geo_alpha_2)

        # Iterate over source attributes
        for av in observation.attached_attribute.values():
            ATTRIBUTES.add(av.value_for.id)  # DEBUG Track all attributes

            if not av.value_for.id.startswith("Source"):
                continue
            sources[geo_alpha_2].add(av.value)

    # Append a sorted list of GEO to the description
    description_lines.extend(
        ["", "The data contain non-null values for GEO:"] + sorted(geo_all)
    )

    # Finalise description
    mdr.metadata.append(make_ra("DATA_DESCR", "\n".join(description_lines)))

    # List of 1 or more metadata reports to return
    result = [mdr]

    # Generate additional metadata reports for upstream data
    for geo, source in chain(*[product([g], sorted(s)) for g, s in sources.items()]):
        dfd_geo = v21.DataflowDefinition(id=f"{geo}_{hash(dfd.id + source)}")
        mdr_geo = v21.MetadataReport(
            attaches_to=make_tok(dfd_geo), metadata=[ra_comment]
        )

        # Store the source attribute value as either the URL or DATA_PROVIDER
        # reported metadata attribute, depending on whether it is a URL
        mda_id = "URL" if _is_url(source) else "DATA_PROVIDER"
        mdr_geo.metadata.append(make_ra(mda_id, source))

        desc = (
            f"Upstream source(s) for 'ATO:{dfd.id}' (“{measure_concept.name}”) "
            f"data for GEO={geo!r}."
        )
        mdr_geo.metadata.append(make_ra("DATA_DESCR", desc))

        result.append(mdr_geo)

    return result


def fetch(*parts, dry_run: bool = False):
    if dry_run:
        for p in parts:
            print(f"Valid url for GEO={p}: {POOCH.is_available(p)}")
        return

    return list(chain(*[POOCH.fetch(p) for p in parts]))


def format_data_provider(value: str) -> str:
    """Format the ATO “Source” data attribute as TDC ``DATA_PROVIDER`` metadata.

    This makes more explicit how the ATO has handled upstream data.
    """
    if value.startswith("ATO analysis"):
        return value.replace("ATO analysis", "Asian Transport Outlook (ATO) analysis")
    elif value.startswith("Estimated"):
        return value.replace("Estimated", "Asian Transport Outlook (ATO); estimated")
    elif value.startswith("Calculated"):
        return "Asian Transport Outlook (ATO) calculation"
    else:
        return value + "—republished by ATO"


@hookimpl
def get_agencies():
    a = m.Agency(
        id="ADB",
        name="Asian Transport Outlook team at the Asian Development Bank",
        description="""See https://www.adb.org/what-we-do/topics/transport/asian-transport-outlook""",  # noqa: E501
    )

    c1 = m.Contact(name="James Leather", email=["jleather@adb.org"])
    c2 = m.Contact(name="Cornie Huizenga", email=["chuizenga@cesg.biz"])
    c3 = m.Contact(name="Sudhir Gota", email=["sudhirgota@gmail.com"])
    a.contact.extend([c1, c2, c3])

    return (a,)


@hookimpl
def provides():
    return (
        "Codelist=TDCI:CL_ATO_ECONOMY",
        "ConceptScheme=TDCI:CS_ATO_MEASURE",
    )


def prepare(aa: m.AnnotableArtefact) -> Tuple[m.DataSet, Callable]:
    """Prepare an empty data set and associated structures."""
    # Measure identifier and description
    measure_id = (
        str(aa.pop_annotation(id="INDICATOR-ATO-CODE").text)
        .replace("(", "_")
        .replace(")", "")
    )
    measure_name = aa.pop_annotation(id="INDICATOR").text
    measure_desc = aa.pop_annotation(id="DESCRIPTION").text

    # Add to the "Measure" concept scheme
    # TODO avoid duplicating items for e.g. TDC-PAT-001(1), TDC-PAT-001(2)
    c = m.Concept(id=measure_id, name=measure_name, description=measure_desc)
    # TODO Extend an existing code list if converting only 1 or a few data sets
    CS_MEASURE.append(c)

    # Data structure definition with an ID matching the measure
    # NB here we set ADB as the maintainer. Precisely, ADB establishes the data
    #    structure, but TDCI is maintaining the SDMX representation of it.
    ma_args = dict(maintainer=get_agencies()[0], version=VERSION)
    dsd = m.DataStructureDefinition(id=measure_id, **ma_args)
    anno_generated(dsd)

    dfd = m.DataflowDefinition(id=measure_id, structure=dsd, **ma_args)

    pm = m.PrimaryMeasure(id="OBS_VALUE", concept_identity=c)
    dsd.measures.append(pm)

    # Dimensions
    dsd.dimensions.extend(m.Dimension(id=d) for d in ("ECONOMY", "TIME_PERIOD"))

    # Convert annotations to DataAttributes. "NoSpecifiedRelationship" means that the
    # attribute is attached to an entire data set (not a series, individual obs, etc.).
    da = {}  # Store references for use below
    for a in filter(lambda a: a.id != "remark-cols", aa.annotations):
        da[a.id] = m.DataAttribute(id=a.id, related_to=m.NoSpecifiedRelationship())
        dsd.attributes.append(da[a.id])

    _PMR = m.PrimaryMeasureRelationship  # Shorthand

    # Convert remark column labels to DataAttributes. "PrimaryMeasureRelationship" means
    # that the attribute is attached to individual observations.
    for name in aa.eval_annotation("remark-cols"):
        dsd.attributes.append(m.DataAttribute(id=name, related_to=_PMR()))

    # Empty data set structured by this DSD
    ds = m.DataSet(
        described_by=dfd,
        structured_by=dsd,
        annotations=[aa.get_annotation(id="remark-cols")],
    )
    anno_generated(ds)

    # Convert temporary annotation values to attributes attached to the data set
    for a in filter(lambda a: a.id != "remark-cols", aa.annotations):
        ds.attrib[a.id] = m.AttributeValue(value=str(a.text), value_for=da[a.id])

    def _make_obs(row):
        """Helper function for making :class:`sdmx.model.Observation` objects."""
        key = dsd.make_key(m.Key, row[[d.id for d in dsd.dimensions]])

        # Attributes
        attrs = {}
        for a in filter(lambda a: isinstance(a.related_to, _PMR), dsd.attributes):
            # Only store an AttributeValue if there is some text
            value = row[a.id]
            if not pd.isna(value):
                attrs[a.id] = m.AttributeValue(value_for=a, value=str(value))

        return m.Observation(
            dimension=key, attached_attribute=attrs, value_for=pm, value=row[pm.id]
        )

    return ds, _make_obs


def read_sheet(
    ef: pd.ExcelFile, sheet_name: str
) -> Tuple[pd.DataFrame, m.AnnotableArtefact]:
    """Read a single sheet.

    This function handles the particular layout of sheets in files like those listed in
    :data:`FILES`. These combine data and metadata.

    - Row 1 is a title row.
    - Cell range A2:B10 contain a set of metadata fields, with the field name in column
      A and the value in column B.
    - Rows 11:13 contain no data or metadata; only a link back to a table of contents
      sheet.
    - Row 14 contains a label "Series" centre-spanned across
    - Row 15 contains column labels, described below.
    - Row 16 and onwards contain data, followed by two blank rows, and two rows with
      attribution/acknowledgements.

    - Columns labeled (i.e. in row 15) "Economy Code" and "Economy Name" contain codes
      and names, respectively, for the geographic units.
    - Columns with numeric labels describe time periods, specifically years, that are
      part of observation keys.
    - Some sheets have additional columns with non-numeric labels like "Remarks",
      "Source (2022-04)", etc.; these give annotations applying to the observations on
      the same row (i.e. for a single "Economy Code" and 1 or more time periods).

    .. note:: Sheets in the ``POL`` category have a different format.
    """
    # Read metadata section
    meta_df = ef.parse(
        sheet_name, header=None, skiprows=1, nrows=9, usecols="A:B"
    ).dropna(how="all")

    # Convert metadata rows to a collection of temporary annotations
    aa = m.AnnotableArtefact()
    for _, data in meta_df.iterrows():
        title = data.iloc[0].rstrip(":")
        anno_id = title.upper().replace(" ", "-")
        aa.annotations.append(
            m.Annotation(
                id=anno_id,
                title=title,
                text=data.iloc[1]
                if isinstance(data.iloc[1], str)
                else repr(data.iloc[1]),
            )
        )

    # Read data section
    df = ef.parse(sheet_name, skiprows=14, skipfooter=2).dropna(how="all")

    # Identify data columns: those with numeric labels
    data_col_mask = list(map(str.isnumeric, df.columns))

    # Handle values not parsed by pd.ExcelFile.parse(). Some cells have values like
    # "12,345.6\t", which are not parsed to float.
    #
    # - Strip trailing whitespace.
    # - Remove thousands separators ("," in 2022-10-17 edition; " " in 2024-05-20
    #   edition) and whitespace before the decimal separator ("10860 .6", 2024-05-20
    #   edition).
    # - RSA-RSI-007_1 contains erroneous values like "55.0g", likely intended to be
    #   "55.0%"; correct this.
    # - Replace miscellaneous erroneous values appearing since the 2024-05-20 edition:
    #   - "-", "n/a", "N/Appl." (both indicating no data)
    #   - "_" (likely erroneous)
    #   - "long ton" (erroneous)
    # - Some sheets contain columns with text labels in miscellaneous/undocumented
    #   formats, not supported by this function:
    #   - SEC-SEG-009: "land", "port"
    #   - MIS-SUM-002: "A", "B", "C", "D"
    #
    #   TODO Ask ATO to provide these data as SDMX, or extend code to convert
    # - Finally, convert to float.
    dtypes = df.loc[:, data_col_mask].dtypes  # Dtypes of data columns only
    for col, _ in filter(lambda x: x[1] != "float", dtypes.items()):
        df[col] = (
            df[col]
            .str.strip()
            .str.replace(r"(\d)[, ]([\d\.])", r"\1\2", regex=True)
            .str.replace(r"^([\d\.]+)g", r"\1", regex=True)
            .str.replace(r"^(-|long ton|N/Appl\.|n/a|_)$", "NaN", regex=True)
            .str.replace(r"^(land|port|A|B|C|D)$", "NaN", regex=True)
            .astype(float)
        )

    # Identify remarks columns: any entries at the *end* of `df.columns` with non-
    # numeric labels. Use the index of last data column, counting backwards.
    try:
        N = list(reversed(data_col_mask)).index(True)
    except ValueError:
        N = 0

    # Label(s) of any remarks columns
    remark_cols = df.columns.tolist()[-N:] if N else []
    # Store a list of these as another annotation
    aa.annotations.append(m.Annotation(id="remark-cols", text=repr(remark_cols)))

    return df, aa


def validate_economy(df: pd.DataFrame) -> pd.DataFrame:
    """Validate codes for the "ECONOMY" dimension of `df` against :data:`CL_ECONOMY`.

    - Every unique pair of (Economy Code, Economy Name) is converted to a
      :class:`~sdmx.model.common.Code`.
    - These are added to :data:`CL_ECONOMY`. If a Code with the same ID already exists,
      it is checked for an exact match (name, description, etc.)
    - The "Economy Code" column of `df` is renamed "ECONOMY", and contains only values
      from :data:`CL_ECONOMY`. The "Economy Name" column is dropped.
    """
    codes = (
        df[["Economy Code", "Economy Name"]]
        .sort_values("Economy Code")
        .drop_duplicates()
        .apply(
            lambda row: m.Code(id=row["Economy Code"], name=row["Economy Name"]), axis=1
        )
    )
    for c in codes:
        try:
            CL_ECONOMY.append(c)
        except ValueError:
            assert CL_ECONOMY[c.id].compare(
                c
            ), f"Existing {CL_ECONOMY[c.id]} does not match {c}"

    return df.rename(columns={"Economy Code": "ECONOMY"}).drop(columns=["Economy Name"])



"""CLI for :mod:`.adb`.

.. runblock:: console

   $ tdc adb --help

.. runblock:: console

   $ tdc add fetch --help

.. runblock:: console

   $ tdc adb convert --help
"""

import click

from . import FILES, convert, fetch


@click.group("adb")
def main():
    """Asian Development Bank (ADB) provider."""


@main.command("fetch")
@click.argument("part", nargs=-1)
@click.option("--go", is_flag=True, help="Actually fetch.")
@click.option("--all", "all_", is_flag=True, help="Fetch all files.")
def fetch_cmd(part, all_, go):
    """Fetch original data files."""
    if not len(part):
        if not all_:
            print(f"Supply --all or 1+ of: {' '.join(FILES)}")
            return

        part = list(FILES.keys())

    fetch(*part)


@main.command("convert")
@click.argument("part", nargs=-1)
@click.option("--all", "all_", is_flag=True, help="Convert all parts.")
def convert_cmd(part, all_):  # pragma: no cover
    """Convert to SDMX data and structures."""
    if not len(part):
        if not all_:
            print(f"Supply --all or 1+ of: {' '.join(FILES)}")
            return

        part = list(FILES.keys())

    for p in part:
        if p == "POL":
            print(f"Skip PART = {p}; not currently supported")
            continue

        convert(p)



import pytest

from transport_data.oica import convert, get_cl_geo, update_registry


@pytest.mark.parametrize(
    "measure, N_obs",
    (
        pytest.param("PROD", None, marks=pytest.mark.xfail(raises=NotImplementedError)),
        ("SALES", dict(SALES=2664, SALES_GR=1330)),
        ("STOCK", dict(STOCK=462, STOCK_AAGR=230, STOCK_CAP=77)),
        ("STOCK_AAGR", dict(STOCK=462, STOCK_AAGR=230, STOCK_CAP=77)),
    ),
)
def test_convert(measure, N_obs):
    result = convert(measure)

    # GEO codelist written with some elements
    assert 0 < len(get_cl_geo())

    # Results contain the expected number of data sets
    assert len(N_obs) == len(result)

    # Each data set contains the expected number of observations
    observed = {dfd_id.split("DF_OICA_")[-1]: len(ds) for dfd_id, ds in result.items()}
    assert N_obs == observed

    # DEBUG
    # import sdmx
    #
    # for _, ds in result.items():
    #     print(sdmx.to_pandas(ds).to_string())


def test_update_registry():
    update_registry()



""":class:`.ckan.Client` objects to access TDC instances."""

from transport_data.util.ckan import Client

#: Development instance of TDC CKAN. Accessible as of 2024-11-27.
DEV = Client("https://ckan.tdc.dev.datopian.com")

#: Production instance of TDC CKAN. Accessible as of 2024-11-27.
PROD = Client("https://ckan.transport-data.org")

#: Staging instance of TDC CKAN. Not accessible as of 2024-11-27; SSL certificate error.
STAGING = Client("https://ckan.tdc.staging.datopian.com")



"""Utilities for :mod:`pooch`."""

from typing import Callable, Optional

import pooch


class Pooch(pooch.Pooch):
    """:class:`pooch.Pooch` with special powers.

    Parameters
    ----------
    module : str
        The `path` argument to pooch.Pooch is derived automatically from this, within
        the transport-data directory.
    expand : callable, optional
        If given, a function that receives any argument and returns a file name found
        within the Pooch registry.
    processor :
        If the exact value "unzip", fetch() operations are always processed to unzip
        archive contents into the same directory as the downloaded file.
    """

    _processor = None

    def __init__(
        self,
        *args,
        module: str,
        expand: Optional[Callable] = None,
        processor=None,
        **kwargs,
    ):
        kwargs.setdefault(
            "path", pooch.os_cache("transport-data").joinpath(module.split(".")[1])
        )

        self._expand = expand or str

        if processor == "unzip":
            self._processor = pooch.Unzip(extract_dir=kwargs["path"])

        super().__init__(*args, **kwargs)

    def fetch(self, fname, *args, **kwargs):
        kwargs.setdefault("processor", self._processor)
        return super().fetch(self._expand(fname), *args, **kwargs)

    def is_available(self, fname, *args, **kwargs):
        return super().is_available(self._expand(fname), *args, **kwargs)

    def path_for(self, *args, **kwargs):
        """Return a filename and local cache path for the data file."""
        return self.path.joinpath(self._expand(*args, **kwargs))



import pytest
from requests.exceptions import SSLError

from transport_data.org.ckan import DEV, PROD, STAGING


@pytest.mark.parametrize(
    "instance, N_exp",
    [
        # (DEV, 451),
        (DEV, 5),
        # (PROD, 418),
        (PROD, 5),
        pytest.param(STAGING, None, marks=pytest.mark.xfail(raises=SSLError)),
    ],
)
def test_main(instance, N_exp: int) -> None:
    result = instance.package_list(max=5)

    assert N_exp <= len(result)

    p = instance.package_show(result[0])

    assert p is result[0]  # Same object instance returned

    assert "dataset" == p.type



"""Utilities for :mod:`sdmx`."""

import io
from datetime import datetime
from importlib.metadata import version
from typing import TYPE_CHECKING, Optional, Union, cast

import pandas as pd

if TYPE_CHECKING:
    import pathlib
    from typing import TypedDict

    import sdmx.model.common
    import sdmx.model.v21
    import sdmx.model.v30

    # Types for dicts holding keyword arguments to SDMX classes
    # TODO Move upstream
    class IAKeywords(TypedDict, total=False):
        id: str

    class VAKeywords(IAKeywords):
        version: str

    class MAKeywords(VAKeywords):
        maintainer: Optional[sdmx.model.common.Agency]


class CSVAdapter(io.RawIOBase):
    """Adapt CSV content from `path` into SDMX-CSV.

    The `SDMX-CSV <https://github.com/sdmx-twg/sdmx-csv/blob/master/data-message/docs/sdmx-csv-field-guide.md>`_
    format has precise requirements for identifier and other columns. In particular, the
    "STRUCTURE", "STRUCTURE_ID", and "ACTION" columns are mandatory.

    .. code-block::

       STRUCTURE,STRUCTURE_ID,ACTION,DIM_1,DIM_2,DIM_3,OBS_VALUE
       dataflow,ESTAT:NA_MAIN(1.6.0),I,A,B,2014-01,12.4

    This class produces standard SDMX-CSV by adapting a simplified format such as:

    .. code-block::

       DIM_1,DIM_2,DIM_3,OBS_VALUE
       A,B,2014-01,12.4

    In particular:

    - The header line of `path` is prefixed with "STRUCTURE", "STRUCTURE_ID", and/or
      "ACTION" if the respective parameters are given.
    - Every record (line) of `path` is prefixed with the *values* of the respective
      parameters.

    So, for instance, to transform the above example into the first example, use a
    CSVAdapter with:

    .. code-block:: python

       a = CSVAdapter(
           path,
           structure="dataflow",
           structure_id="ESTAT:NA_MAIN(1.6.0)",
           action="I",
       )
       data = a.read()

    Parameters
    ----------
    structure :
        Value for the "STRUCTURE" SDMX-CSV field, to be inserted into every record.
    structure_id :
        Value for the "STRUCTURE_ID" field, to be inserted into every record.
    action :
        Value for the "ACTION" field, to be inserted into every record.
    """

    def __init__(
        self,
        path: "pathlib.Path",
        structure: Optional[str] = None,
        structure_id: Optional[str] = None,
        action: Optional[str] = None,
    ) -> None:
        self._path = path

        # Determine fields to prefix to header line and records
        header, line = [], []
        for column, value in (
            (b"STRUCTURE", structure),
            (b"STRUCTURE_ID", structure_id),
            (b"ACTION", action),
        ):
            if value is not None:
                header.append(column)
                line.append(value.encode())

        # Construct single `bytes`` for header and line prefixes
        self._header_prefix = b",".join((header + [b""]) if len(header) else [])
        self._line_prefix = b",".join((line + [b""]) if len(line) else [])

    def readall(self, size: int = -1, /) -> bytes:
        """Read and adapt CSV to SDMX-CSV.

        .. todo:: This currently reads and adapts the entire file at once; this may not
           be performant for very large files. Improve to handle lines individually or
           in batches.
        """
        lines = []
        with open(self._path, "rb") as f:
            lines = [
                (self._header_prefix if i == 0 else self._line_prefix) + line
                for i, line in enumerate(f)
            ]

        return b"".join(lines)


def anno_generated(obj: "sdmx.model.common.AnnotableArtefact") -> None:
    """Annotate the `obj` with information about how it was generated."""
    from sdmx.model import v21 as m

    try:
        # Retrieve existing annotation
        anno = obj.get_annotation(id="tdc-generated")
    except KeyError:
        # Create a new annotation
        anno = m.Annotation(id="tdc-generated")
        obj.annotations.append(anno)

    anno.text = (
        f"{datetime.now().isoformat()} by transport_data v{version('transport_data')}"
    )


def make_obs(
    row: "pd.Series", dsd: "sdmx.model.v21.DataStructureDefinition"
) -> "sdmx.model.v21.Observation":
    """Helper function for making :class:`sdmx.model.Observation` objects."""
    from sdmx.model import v21 as m

    key = dsd.make_key(m.Key, row[[d.id for d in dsd.dimensions]].to_dict())

    # Attributes
    attrs = {}
    for a in filter(
        lambda a: isinstance(a.related_to, m.PrimaryMeasureRelationship), dsd.attributes
    ):
        # Only store an AttributeValue if there is some text
        value = row[a.id]
        if not pd.isna(value):
            attrs[a.id] = m.AttributeValue(value_for=a, value=value)

    pm = dsd.measures[0]
    return m.Observation(
        dimension=key, attached_attribute=attrs, value_for=pm, value=row[pm.id]
    )


def read_csv(
    path: "pathlib.Path",
    structure: Union[
        "sdmx.model.v30.Dataflow", "sdmx.model.v30.DataStructureDefinition"
    ],
    adapt: Optional[dict] = None,
) -> "sdmx.message.DataMessage":
    """Read or adapt SDMX-CSV from `path`.

    Parameters
    ----------
    path :
        A file in SDMX-CSV or CSV format.
    structure :
        Data flow or data structure describing the contents of `path`.
    adapt :
        Keyword arguments to :class:`CSVAdapter`. If given, the contents of `path` are
        adapted from a ‘simplified’ or ‘reduced’ CSV format to SDMX-CSV on-the-fly. See
        the class documentation for details.
    """
    import sdmx

    if adapt:
        source: Union["pathlib.Path", "CSVAdapter"] = CSVAdapter(path, **adapt)
    else:
        source = path

    return cast(
        "sdmx.message.DataMessage",
        sdmx.read_sdmx(source, format="csv", structure=structure),
    )



"""Utilities for interacting with CKAN instances via their Action API.

The :mod:`ckanapi` package (`PyPI <https://pypi.org/project/ckanapi/>`_,
`GitHub <https://github.com/ckan/ckanapi>`_), maintained by the CKAN organization,
provides Pythonic access to the “CKAN Action API” (
`documentation <https://docs.ckan.org/en/latest/api/index.html>`_), which itself exposes
nearly all of the functionality available through the CKAN web interface and the TDC
front-end.

The :class:`~.ckan.Client` class is a wrapper around the :class:`ckanapi.RemoteCKAN`
class that provides conveniences used by other code in :mod:`transport_data`.
"""

from functools import partialmethod
from importlib.metadata import version
from itertools import count
from typing import TYPE_CHECKING, Optional, TypeVar, Union
from warnings import filterwarnings

if TYPE_CHECKING:
    import pathlib

    from ckanapi import RemoteCKAN

# Work around https://github.com/ckan/ckanapi/pull/218 in ckanapi <= 4.8
filterwarnings(
    "ignore", "pkg_resources is deprecated", DeprecationWarning, "ckanapi.version"
)
filterwarnings(
    "ignore", ".*pkg_resources.declare_namespace", DeprecationWarning, "pkg_resources"
)

T = TypeVar("T", bound="ModelProxy")


class ModelProxy:
    """Simple proxy for a CKAN object/model.

    :mod:`ckan` itself is a Python package, but is fairly ‘heavy’—a large package with
    many dependencies. ModelProxy allows to interact with the different classes of
    CKAN objects based on the JSON data returned by the CKAN Action API, without a
    dependency on :class:`ckan` itself.
    """

    name: Optional[str] = None
    id: Optional[str] = None

    _collections: dict[str, str] = dict()

    def __init__(self, data: Optional[dict] = None, **kwargs) -> None:
        self.__dict__.update(data or {})
        self.__dict__.update(kwargs)

    def __len__(self) -> int:
        return len(self.__dict__)

    def __repr__(self) -> str:
        return (
            f"<CKAN {type(self).__name__} "
            + (repr(self.name) if self.name else "(no name)")
            + f" with {len(self.__dict__) - 1} fields>"
        )

    @classmethod
    def from_file(cls: type[T], path: "pathlib.Path") -> T:
        """Construct a new instance from a file `path`."""
        import json

        with open(path) as f:
            return cls(json.load(f))

    def asdict(self) -> dict:
        """Return the original dictionary of object data."""
        return self.__dict__.copy()

    def get_item(self, name: str, index: Optional[int] = None):
        """Get a member of a collection."""
        data = self.__dict__[name][index]
        cls = get_class(name)
        assert cls
        return cls(data)

    def update(self, data: dict) -> None:
        """Update part or all of the object data."""
        # Check some items
        if self.name and data.get("name", self.name) != self.name:
            raise ValueError(f"Cannot update with {data['name']!r} != {self.name=!r}")
        elif self.id and data.get("id", self.id) != self.id:
            raise ValueError(f"Cannot update with {data['id']!r} != {self.id=!r}")
        self.__dict__.update(data)


def get_class(name: str) -> Optional[type[ModelProxy]]:
    """Return a :class:`.ModelProxy` subclass given `name`."""
    glb = globals()
    for candidate in (
        name.title(),
        name.rstrip("s").title(),
        {"member_role": "MemberRole"}.get(name.rstrip("s"), ""),
    ):
        try:
            return glb[candidate]
        except KeyError:
            pass
    return None


class Group(ModelProxy):
    """Proxy for `ckan.model.Group
    <https://github.com/ckan/ckan/blob/master/ckan/model/group.py>`_.
    """


class Organization(Group):
    """'Organization' is a synonym for 'Group'."""

    # NB this is a subclass instead of `Organization = Group` so that type(…).__name__
    #    gives 'organization'


class MemberRole(ModelProxy):
    """Proxy for the CKAN 'MemberRole' model.

    .. todo:: Add a link to the proxied class.
    """


class License(ModelProxy):
    """Proxy for `ckan.model.License
    <https://github.com/ckan/ckan/blob/master/ckan/model/license.py>`_.
    """


class Package(ModelProxy):
    """Proxy for `ckan.model.Package
    <https://github.com/ckan/ckan/blob/master/ckan/model/package.py>`_.
    """


class Resource(ModelProxy):
    """Proxy for `ckan.model.Resource
    <https://github.com/ckan/ckan/blob/master/ckan/model/resource.py>`_.
    """


class Tag(ModelProxy):
    """Proxy for the CKAN 'Tag' model.

    The source code for the proxied class is `here
    <>`_.
    """


class Client:
    """Wrapper around :class:`ckanapi.RemoteCKAN`.

    This provides features such as:

    - Iterating over calls with rate limits.
    - Caching results, including combined results from multiple calls.
    - Converting return values to instances of :class:`ModelProxy` subclasses.

    .. todo::
       - Handle API keys via :mod:`transport_data.config`.
    """

    _api: "RemoteCKAN"
    _cache: dict

    def __init__(self, address: str) -> None:
        from ckanapi import RemoteCKAN

        # Construct a user-agent string
        user_agent = (
            f"transport_data/{version('transport_data')} "
            "(+https://docs.transport-data.org)"
        )

        self._api = RemoteCKAN(address, user_agent=user_agent)
        self._cache = dict(package=dict())

    def __getattr__(self, name: str):
        return getattr(self._api.action, name)

    def list_action(
        self, kind: str, limit: Optional[int] = None, max: Optional[int] = None
    ) -> list:
        """Call the ``{kind}_list`` API endpoint.

        Parameters
        ----------
        kind
            String identifying the kind of CKAN object to fetch.
        limit
            Number of results to fetch in a single query.
        max
            Maximum number of results to fetch.
        """
        cls = get_class(kind)
        assert cls

        limit = limit or 100

        c = self._cache[f"{kind}_list"] = list()

        for i in count():
            result = self._api.call_action(
                f"{kind}_list", data_dict={"limit": limit, "offset": i * limit}
            )
            c.extend(result)
            if len(result) < limit or (max and max < (i + 1) * limit):
                break
        return [(cls(name=v) if isinstance(v, str) else cls(v)) for v in c]

    # Use list_action() to invoke certain CKAN API endpoints
    group_list = partialmethod(list_action, "group")
    license_list = partialmethod(list_action, "license")
    member_roles_list = partialmethod(list_action, "member_roles")
    organization_list = partialmethod(list_action, "organization")
    package_list = partialmethod(list_action, kind="package")
    tag_list = partialmethod(list_action, kind="tag")

    def show_action(self, obj_or_id: Union[str, dict, T], _cls: type[T]) -> T:
        """Call the ``{kind}_show`` API endpoint.

        If `obj_or_id` is an instance of :class:`.ModelProxy`, its
        :attr:`~ModelProxy.name` or :attr:`~.ModelProxy.id` is used for the query, and
        the same instance is updated with the response data and returned. If `obj_or_id`
        is :class:`str` or :class:`dict` a new instance of a :class:`.ModelProxy`
        subclass is returned.

        The return value is cached.
        """
        if isinstance(obj_or_id, dict):
            kw: dict[str, str] = obj_or_id
            kw.update(id=kw.pop("name", kw.get("id", "")))
            result: Optional[T] = _cls()
        else:
            kw = dict()

        if isinstance(obj_or_id, str):
            # Query using the "id" keyword
            kw.update(id=obj_or_id)
            # Create a new result object according to the type of
            result = None
        elif isinstance(obj_or_id, _cls):
            if obj_or_id.id:
                kw.update(id=obj_or_id.id)
            elif obj_or_id.name:
                kw.update(id=obj_or_id.name)
            result = obj_or_id

        kind = _cls.__name__.lower()

        response = self._api.call_action(f"{kind}_show", kw)

        # TODO Handle whatever exception is raised for an invalid `package`
        if result is None:
            result = _cls(response)
        else:
            result.update(response)

        # Update cache by both UUID and name
        self._cache[result.id] = result
        self._cache.setdefault(kind, dict())
        self._cache[kind][result.name] = result

        return result

    # Use show_action() to invoke certain CKAN API endpoints
    # TODO Extend to cover all "*_show" endpoints.
    organization_show = partialmethod(show_action, _cls=Organization)
    package_show = partialmethod(show_action, _cls=Package)
    tag_show = partialmethod(show_action, _cls=Tag)
